package celery.engine;

import java.util.ArrayList;

import android.content.Context;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.View;
import celery.engine.graphics.SceneEntity;
import celery.engine.graphics.Viewport;
import celery.engine.inputhandlers.KeyboardInputHandler;
import celery.engine.inputhandlers.TouchInputHandler;
import celery.engine.managers.AudioManager;
import celery.engine.managers.BitmapManager;
import celery.engine.managers.MaterialManager;
import celery.engine.updatehandlers.DrawHandler;
import celery.engine.updatehandlers.PhysicsHandler;
import celery.engine.updatehandlers.UpdateHandler;
import celery.engine.utils.MessageHandler;
import celery.engine.utils.math.Vec2F;

/**
 * Celery Android Game Engine
 * 
 * The base 2D game engine. This is a singleton class which is automatically
 * instantiated by the engine activity. It's instance can be accessed by calling
 * Engine.instance() to retrieve the singleton instance.
 * 
 * Keep in mind that the settings you initialize the engine with persist
 * throughout the life of the engine, and cannot be changed (so if you want to
 * use sound, you must initiate the engine with the appropriate settings - see
 * EngineActivity for more information).
 * 
 * @author psastras
 * 
 */
public final class Engine {

	// / Engine settings holder which should be constant throughout the life of
	// the engine
	protected EngineSettings settings_;
	// / Container holding all of the update handlers (which receive
	// notifications)
	protected ArrayList<UpdateHandler> updateHandlers_;
	// / Message handler which sends and receives messages
	protected MessageHandler msgHandler_;
	// / Draw handler which contains the surface and performs most drawing
	protected DrawHandler drawHandler_;
	// / Physics handler which is a specialized handler responsible for physics
	protected PhysicsHandler physicsHandler_;
	// / Touch input handler which is used to handle all screen touch inputs
	protected TouchInputHandler touchHandler_;
	// / Keyboard input handler which is used to handle all key events
	protected KeyboardInputHandler keyboardHandler_;
	// / The context associated with the engine (this should be the same as the
	// root application context)
	protected Context context_;
	// / The start system time (in msecs) of the engine
	protected long startTime_;
	// / Whether or not the engine is running (controlled by
	// start/stop/pause/resume)
	protected boolean isRunning_;

	/**
	 * Private engine constructor.
	 * 
	 * @param context
	 *            The main application context.
	 * @param settings
	 *            The engine settings to initialize the engine with.
	 */
	private Engine(Context context, EngineSettings settings) {
		settings_ = settings;

		context_ = context;
		updateHandlers_ = new ArrayList<UpdateHandler>();
		msgHandler_ = new MessageHandler();
		isRunning_ = false;
		this.applySettings(settings);

	}

	public long startTime() {
		return startTime_;
	}

	/**
	 * Pauses the game engine as well as all of its associated update handlers.
	 * Each update handler should be responsible for pausing itself (ex.
	 * notifications) The games running state is set to false (see isRunning_)
	 */
	public void pause() {
		for (UpdateHandler eh : updateHandlers_)
			eh.pause();

		isRunning_ = false;
	}

	/**
	 * Resumes the game engine as well as all of its associated update handlers.
	 * Each update handler should be responsible for resuming itself (Ex.
	 * notifcations) The game's running state is set to true (see isRunning_),
	 * and the previous startTime_ is reset to the current system time in
	 * milliseconds.
	 */
	public void resume() {
		startTime_ = System.currentTimeMillis();
		isRunning_ = true;
		for (UpdateHandler eh : updateHandlers_)
			eh.resume();
	}

	/**
	 * Starts running the engine. This method should only be called once to
	 * restart a paused engine, use resume, which will skip the engine check. If
	 * you are extending / implementing EngineActivity, you should not need to
	 * call this method.
	 */
	public void start() {
		startTime_ = System.currentTimeMillis();
		isRunning_ = true;
		this.check();
		for (UpdateHandler eh : updateHandlers_)
			eh.start();

	}

	/**
	 * Checks the engine settings and provided handlers as well as systems specs
	 * to ensure that the engine settings are compatible with the current setup.
	 * If the settings are not compatible, errors will be output, but the engine
	 * will try to continue setup anyway - although this may result in weird
	 * behavior.
	 */
	protected void check() {
		if (settings_.useKeyboardInput && keyboardHandler_ == null) {
			System.err.println("Keyboard handler requested, but not set.");
		}

		if (settings_.useTouchInput && touchHandler_ == null) {
			System.err.println("Touch handler requested, but not set.");
		}
	}

	/**
	 * Returns the instance of the DrawHandler associated with the engine.
	 * 
	 * @return The DrawHandler instance.
	 */
	public DrawHandler drawHandler() {
		return drawHandler_;
	}

	/**
	 * Returns the instance of the PhysicsHandler associated with the engine.
	 * 
	 * @return The PhysicsHandler instance.
	 */
	public PhysicsHandler physicsHandler() {
		return physicsHandler_;
	}

	/**
	 * Returns the elapsed time of the engine since the last start() or paus()
	 * in milliseconds.
	 * 
	 * @return Elapsed time in milliseconds.
	 */
	public long elapsedTime() {
		return System.currentTimeMillis() - startTime_;
	}

	/**
	 * Applies the current engine settings to the engine. This should only be
	 * called once. The DrawHandler, ImageManager, and PhysicsManger, as well as
	 * other Engine instance variables are instantiated here.
	 * 
	 * @param settings
	 *            The EngineSettings to initialize to.
	 */

	private SceneEntity selected_ = null;

	protected void applySettings(EngineSettings settings) {
		if (settings.drawSettings != null) {
			BitmapManager.init();
			drawHandler_ = new DrawHandler(settings.drawSettings, context_);
			drawHandler_.setViewport(Viewport.defaultViewport());
			updateHandlers_.add(drawHandler_);
			MaterialManager.init();
		}

		if (settings.physicsSettings != null) {
			physicsHandler_ = new PhysicsHandler(settings.physicsSettings,
					context_);
			updateHandlers_.add(physicsHandler_);
		}

		if (settings.useTouchInput) {
			touchHandler_ = new TouchInputHandler() {

				@Override
				public boolean onTouch(SceneEntity e, View arg0,
						MotionEvent arg1) {
					Vec2F location = new Vec2F(arg1.getX(), arg1.getY());
					if (drawHandler().scene() != null) {
						SceneEntity se = drawHandler().scene()
								.locateSceneObject(
										drawHandler().viewport()
												.screenToVirtual(location));
						selected_ = se;
						if (se != null) {

							if (se.touchInputHandler().isEnabled()) {
								se.onTouch(arg0, arg1);
								return se.touchInputHandler().onTouch(se, arg0,
										arg1);
							}
						}
					}
					return false;
				}

				@Override
				public boolean onDrag(SceneEntity e, View arg0, MotionEvent arg1) {
					if (drawHandler().scene() != null) {
						SceneEntity se = selected_;
						if (se != null) {

							if (se.touchInputHandler().isEnabled()) {
								se.onDrag(arg0, arg1);
								return se.touchInputHandler().onDrag(se, arg0,
										arg1);
							}
						}
					}
					return false;
				}

				@Override
				public boolean onRelease(SceneEntity e, View arg0,
						MotionEvent arg1) {
					if (drawHandler().scene() != null) {
						SceneEntity se = selected_;
						if (se != null) {

							if (se.touchInputHandler().isEnabled()) {
								se.onRelease(arg0, arg1);
								boolean toReturn = se.touchInputHandler()
										.onRelease(se, arg0, arg1);
								selected_ = null;
								return toReturn;
							}
						}
					}
					return false;
				}
			};
		}

		if (!settings.useKeyboardInput) {
			keyboardHandler_ = new KeyboardInputHandler() {

				@Override
				public boolean onKey(View arg0, int arg1, KeyEvent arg2) {
					// TODO Auto-generated method stub
					return false;
				}

			};
		}

		AudioManager.init();
	}

	/**
	 * Static class used to hold the Engine instance used to enforce the
	 * Singleton instance.
	 * 
	 * @author psastras
	 * 
	 */
	private static class EngineInstance {
		public static Engine instance = null;
	}

	/**
	 * Adds an UpdateHandler to the Engine's list of UpdateHandlers. This
	 * UpdateHandler will be paused / resumed when the Engine is paused and/or
	 * resumed.
	 * 
	 * @param handler
	 *            The UpdateHandler to add.
	 */
	public void addUpdateHandler(UpdateHandler handler) {
		updateHandlers_.add(handler);
	}

	/**
	 * Initializes the engine by instantiating / calling the Engine's
	 * constructor. This should only be called once, since it will overwrite the
	 * previous Engine's instance (since this is a singleton class).
	 * 
	 * @param context
	 *            The main application context.
	 * @param settings
	 *            Engine settings to instantiate the Engine with.
	 */
	public static void init(Context context, EngineSettings settings) {
		EngineInstance.instance = new Engine(context, settings);
	}

	/**
	 * Returns the static instance of the Engine. You should use this method
	 * whenever you need to access any of the Engine's members (methods).
	 * 
	 * @return Returns the singleton instance of the Engine.
	 */
	public static Engine instance() {
		return EngineInstance.instance;
	}

	/**
	 * Returns the MessageHandler class associated with the Engine.
	 * 
	 * @return The MessageHandler;
	 */
	public MessageHandler messageHandler() {
		return msgHandler_;
	}

	/**
	 * Returns the ImageManager class associated with the Engine. Note that
	 * since the ImageManager is itself a singleton class, the ImageManager can
	 * be accessed by simply calling ImageManager.instance() from any other
	 * class.
	 * 
	 * @return The ImageManager instance.
	 */
	public BitmapManager imageManager() {
		return BitmapManager.instance();
	}

	/**
	 * Returns the TouchHandler instance associated with the Engine. The touch
	 * handler is responsible for handling all ui touch input events fired by a
	 * Surface.
	 * 
	 * @return The TouchInputHandler instance.
	 */
	public TouchInputHandler touchHandler() {
		return touchHandler_;
	}

	/**
	 * Returns the KeyboardInputHandler instance associated with the Engine. The
	 * keyboard handler is responsible for handling all ui keyboard input events
	 * fired by a Surface.
	 * 
	 * @return The KeyboardInputHandler instance.
	 */
	public KeyboardInputHandler keyboardHandler() {
		return keyboardHandler_;
	}

	/**
	 * Returns whether or not the Engine is currently running. This is false if
	 * pause() was last called, or if start() has not yet been called, and true
	 * otherwise, or if resume() / start() has been called last.
	 * 
	 * @return Whether the Engine is currently running.
	 */
	public boolean isRunning() {
		return isRunning_;
	}

}
