package celery.engine;

import android.app.Activity;
import android.content.pm.ActivityInfo;
import android.os.Bundle;
import android.view.Window;
import android.view.WindowManager;

/**
 * The EngineActivity is an instance of the Android.app.Activity, and is the
 * main Activity of the Application. It is responsible for setting up the Engine
 * and calling the appropriate initialization methods. When writing an
 * application, it should subclass the EngineActivity as its main Activity. At
 * certain points during Engine initialization, the EngineActivity will call
 * certain methods, which can be reimplemented if desired. However, any subclass
 * must reimplement onStartup() which should return the EngineSettings which
 * will be used to instantiate the Engine.
 * 
 * @author psastras
 * 
 */

public abstract class EngineActivity extends Activity {
	@Override
	/**
	 * This method is reimplemented from the Activity base class.  This is where
	 * the Engine is instantiated and initialized the the EngineSettings returned 
	 * by onStartup().
	 */
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		EngineSettings settings = this.onStartup();
		Engine.init(this, settings);
		this.setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_LANDSCAPE);
		this.requestWindowFeature(Window.FEATURE_NO_TITLE);
		this.getWindow().setFlags(WindowManager.LayoutParams.FLAG_FULLSCREEN,
				WindowManager.LayoutParams.FLAG_FULLSCREEN);

		setContentView(Engine.instance().drawHandler());

		this.onEngineFinalized();

		Engine.instance().start();
		this.onEngineStarted();
	}

	/**
	 * Called before the engine has been created. Should return the Engine
	 * settings you wish to initialize the engine with.
	 * 
	 * @return EngineSettings to initilize the Engine with.
	 */
	public abstract EngineSettings onStartup();

	/**
	 * Called after the engine has been initialized. Resources and engine
	 * logical handlers should be initialized here. Handlers can be added to the
	 * Engine by using the addHandler() method.
	 * 
	 */
	public void onEngineFinalized() {

	}

	/**
	 * This method is called after the Engine has been started and is running.
	 * This method does nothing, but can be reimplemented to do something useful
	 * once the Engine has started running.
	 */
	public void onEngineStarted() {

	}

	@Override
	protected void onPause() {
		// @TODO add state saving
		Engine.instance().pause();
		super.onPause();
	}

	@Override
	protected void onResume() {
		Engine.instance().resume();
		super.onResume();
	}

}