package celery.engine.graphics;

import java.util.concurrent.CopyOnWriteArrayList;

import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import celery.engine.Engine;
import celery.engine.utils.math.Vec2F;
import celery.engine.utils.math.Vec4F;

/**
 * Scene which contains a set number of SceneEntities. The Scene is what
 * contains all of the elements to be drawn at a time. The SceneEntities are
 * stored in a SceneGraph with a SceneNode as the root node, where
 * transformations are composed to find the child transformations. Multiple
 * scenes may be created (and probably should), but a DrawHandler can only have
 * one scene set at a time.
 **/
public class Scene {

	protected CopyOnWriteArrayList<SceneEntity> scene_ = new CopyOnWriteArrayList<SceneEntity>();
	protected CopyOnWriteArrayList<Animation> animations_ = new CopyOnWriteArrayList<Animation>();
	protected CopyOnWriteArrayList<SceneEntity> touchables_ = new CopyOnWriteArrayList<SceneEntity>();
	// / Specifies whether the scene is dirty or not. A scene is dirty if
	// objects have been added or removed since the last draw.
	protected boolean isSceneDirty_ = true;
	protected int backgroundColor_ = Color.WHITE;
	protected Bitmap backgroundBitmap_ = null;

	/**
	 * The scene constructor. Currently doesn't do anything. A scene must be
	 * created by adding items to the root node.
	 */
	public Scene() {
		// root_ = new SceneNode(root, null);
	}

	public void clear() {
		scene_.clear();
		animations_.clear();
		touchables_.clear();
	}

	public CopyOnWriteArrayList<SceneEntity> items() {
		return scene_;
	}

	public Bitmap backgroundImage() {
		return backgroundBitmap_;
	}

	public int backgroundColor() {
		return backgroundColor_;
	}

	public void setBackgroundBitmap(Bitmap b) {
		backgroundBitmap_ = b;
	}

	public void setBackgroundColor(int color) {
		backgroundColor_ = color;
	}

	/**
	 * Responsible for calling the draw methods of all the entities contained
	 * within the scene that should appear on the screen. If the root node is
	 * null, this returns, doing nothing. Otherwise, it checks if the scene is
	 * dirty (objects have been added or removed since the last draw), and if it
	 * is dirty, it prepares the scene for drawing by zsorting all of the scene
	 * entities, and culling objects outside of the viewport. Finally, it draws
	 * all of the SceneEntities.
	 * 
	 * @param c
	 *            The application drawing canvas.
	 * @param e
	 *            The Engine instance.
	 * @param v
	 *            The current viewport.
	 */
	Vec4F bounds = new Vec4F();

	public void draw(final Canvas c, final Engine e, final Viewport v) {
		if (v == null)
			return;

		if (isSceneDirty_)
			this.prepareSceneForDraw();

		synchronized (animations_) {
			for (Animation a : animations_) {
				if (a.isFinished())
					animations_.remove(a);
				else if (!a.isRunning())
					a.start();
				else
					a.runTimeStep();
			}
		}

		synchronized (scene_) {

			int length = scene_.size();
			c.save();
			c.translate(v.offset().x, v.offset().y);
			c.scale(v.ratio(), v.ratio());

			for (int i = 0; i < length; i++) {
				SceneEntity se = scene_.get(i);
				if (se.isMarkedForRemoval()) {
					scene_.remove(i);
					i--;
					length--;
					continue;
				}
				// final Vec2F l = v.virtualToScreen(se.transform().translate);
				// final Vec2F r = v.virtualToScreen(se.transform().scale.add(se
				// .transform().translate));
				// bounds.x = l.x;
				// bounds.y = l.y;
				// bounds.z = r.x;
				// bounds.w = r.y;

				c.save();
				se.pushTransform(c);
				se.draw(c, e, v, bounds);
				c.restore();
			}
			c.restore();
		}

	}

	/**
	 * Adds an entity to the scene. When adding entities, touch and physics
	 * handlers should be already applied, otherwise they may not necessarily
	 * have any effect.
	 * 
	 * @param e
	 *            The SceneEntity to add.
	 */
	public void addEntity(SceneEntity e) {
		isSceneDirty_ = true;
		synchronized (scene_) {
			scene_.add(e);
		}

		if (e.hasTouchInput())
			touchables_.add(e);
	}

	protected void prepareSceneForDraw() {
		// do zorder, and potentially cull things
		isSceneDirty_ = false;
	}

	public void addAnimation(Animation a) {

		synchronized (animations_) {
			animations_.add(a);
		}
	}

	public CopyOnWriteArrayList<SceneEntity> touchables() {
		return touchables_;
	}

	public SceneEntity locateSceneObject(Vec2F virtualCoords) {
		int l = touchables_.size() - 1;
		for (int i = l; i >= 0; --i) {
			SceneEntity se = touchables_.get(i);
			if (se.isInBounds(virtualCoords)) {
				return se;
			}
		}
		return null;
	}
	/*
	 * public Entity removeEntity(SceneEntity e) {
	 * 
	 * }
	 */

	/**
	 * The SceneNode class holds the Node's SceneEntity, as well as a reference
	 * to its parent and all of its children.
	 * 
	 * @author psastras
	 * 
	 */
	/*
	 * protected class SceneNode {
	 * 
	 * public SceneNode(SceneEntity e, SceneNode parent) { parent_ = parent;
	 * entity_ = e; children_ = new ArrayList<Scene.SceneNode>(); }
	 * 
	 * // A list of all of the node's children private final
	 * ArrayList<SceneNode> children_; // The SceneEntity associated with this
	 * node SceneEntity entity_; // The parent node of this node. Note that for
	 * a root node, this is // null. SceneNode parent_; }
	 */

}
