package celery.engine.graphics;

import java.util.Collection;

import android.graphics.Canvas;
import android.view.MotionEvent;
import android.view.View;
import celery.engine.Engine;
import celery.engine.graphics.entities.EntityType;
import celery.engine.graphics.entities.GeometryData;
import celery.engine.graphics.entities.Touchable;
import celery.engine.inputhandlers.TouchInputHandler;
import celery.engine.utils.EngineConstants;
import celery.engine.utils.SpatialEntity;
import celery.engine.utils.math.Transform;
import celery.engine.utils.math.Vec2F;
import celery.engine.utils.math.Vec4F;

/**
 * A SceneEntity is a drawable object which can be added to a Scene.
 * SceneEntities may be shared within Scenes, but keep in mind that its actual
 * final transformation in a Scene may be different depending on the previous
 * parent transformations in the Scene's graph.
 * 
 * By default, a SceneEntity is assumed not to have any animation or touch
 * handler.
 * 
 * @author psastras
 * 
 */
public abstract class SceneEntity implements SpatialEntity, Touchable {

	// / The transformation associated with this Entity (scale, rotate, and
	// translate)
	protected Transform transfo_;
	protected TouchInputHandler touchHandler_;
	protected SceneMaterial material_;
	protected boolean markForRemoval_;
	protected Vec2F center_;
	protected EntityType type_;
	protected GeometryData geoDataModel_, geoDataWorld_;

	/**
	 * The scene entity constructor. Note that if the transform is null, it will
	 * automatically be set to the identity transformation (0 rotation, 0
	 * translation and 1 scale). SceneMaterical cannot be null. If
	 * TouchInputHandler is null, this object is assumed to not have any touch
	 * input capabilities, and will not be queried when a touch input is
	 * detected.
	 * 
	 * @param tfo
	 *            The transformation associated with this scene entity.
	 *            (relating the entity's model space with the Viewport's world
	 *            space.
	 * @param mat
	 *            The material associated with this object, which may be shared
	 *            with other entities. Materials should be stored by the
	 *            material manager.
	 * @param tHdl
	 *            The touch handler associated with this entity. The touch
	 *            handler will be fired if this entity detects a touch event.
	 */
	public SceneEntity(EntityType type, Transform tfo, SceneMaterial mat,
			TouchInputHandler tHdl) {
		transfo_ = tfo == null ? EngineConstants.IdentityTransform : tfo;
		material_ = mat;
		touchHandler_ = tHdl;
		markForRemoval_ = false;
		center_ = new Vec2F();
		type_ = type;
		this.setTransform(new Transform(transfo_));
	}

	public EntityType type() {
		return type_;
	}

	public boolean isCollision(SceneEntity e) {
		return this.geometryData().isCollision(e.geometryData());
	}

	public boolean isCollision(Collection<SceneEntity> es) {
		for (SceneEntity se : es) {
			if (isCollision(se))
				return true;
		}
		return false;
	}

	public final Vec2F center() {
		return center_;
	}

	public final boolean hasTouchInput() {
		return touchHandler_ != null;
	}

	public final SceneMaterial material() {
		return material_;
	}

	public final void setMaterial(SceneMaterial mat) {
		material_ = mat;
	}

	@Override
	public final TouchInputHandler touchInputHandler() {
		return touchHandler_;
	}

	public void markForRemoval() {
		markForRemoval_ = true;
	}

	public boolean isMarkedForRemoval() {
		return markForRemoval_;
	}

	/**
	 * Returns the transformation associated with this scene entity.
	 * 
	 * @return Transformation of the Entity.
	 */
	public final Transform transform() {
		return transfo_;
	}

	/**
	 * Sets the current transformation of the scene entity to the new
	 * transformation. The old transformation is lost.
	 * 
	 * @param transform
	 *            The transformation to apply.
	 */
	public final void setTransform(final Transform transform) {
		transfo_ = transform;
		if (transfo_ != null) {
			this.computeCenter(Engine.instance().drawHandler().viewport());
		}
	}

	public void computeCenter(final Viewport v) {
		// center_ = transfo_.translate.add((transfo_.scale).scale(0.5f));
		center_ = transfo_.translate;
		// center_ = v.virtualToScreen(center_);
	}

	/**
	 * The draw method is called by the Scene when drawing the SceneEntities.
	 * This method is reimplemented by other entities for drawing.
	 * 
	 * @param c
	 *            The application main canvas.
	 * @param e
	 *            The engine instance.
	 * @param v
	 *            The viewport to use.
	 */
	public abstract void draw(final Canvas c, final Engine e, final Viewport v,
			final Vec4F bounds);

	public boolean onTouch(View arg0, MotionEvent arg1) {
		return false;
	}

	public boolean onDrag(View arg0, MotionEvent arg1) {
		return false;
	}

	public boolean onRelease(View arg0, MotionEvent arg1) {
		return false;
	}

	public void pushTransform(Canvas c) {
		final Viewport v = Engine.instance().drawHandler().viewport();
		c.save();
		this.pushTranslation(c, v);
		this.pushRotation(c, v);
		this.pushScale(c, v);
	}

	public void popTransform(Canvas c) {
		c.restore();
	}

	public void pushRotation(Canvas c, Viewport v) {
		this.computeCenter(v);
		// if (this.transform().theta != 0.f) {
		// c.rotate(-this.transform().theta, center_.x, center_.y);
		// }
		c.rotate(this.transform().theta);
	}

	public void pushTranslation(Canvas c, Viewport v) {
		c.translate(transfo_.translate.x, transfo_.translate.y);
	}

	public void popTranslation(Canvas c, Viewport v) {
		Vec2F trans = v.virtualToScreen(transfo_.translate);
		c.translate(-trans.x, -trans.y);
	}

	public void pushScale(Canvas c, Viewport v) {
		this.computeCenter(v);
		c.scale(transfo_.scale.x, transfo_.scale.y, 0.f, 0.f);

	}

	public void popScale(Canvas c, Viewport v) {
		this.computeCenter(v);
		c.scale(1.f / transfo_.scale.x / v.ratio(),
				1.f / transfo_.scale.y / v.ratio(), 0.f, 0.f);

	}

	public void popRotation(Canvas c, Viewport v) {
		if (this.transform().theta != 0.f)
			c.rotate(-this.transform().theta);
		// c.rotate(this.transform().theta, center_.x, center_.y);
	}

	public abstract GeometryData geometryData();
}
