package celery.engine.graphics.entities;

import celery.engine.utils.Utils;
import celery.engine.utils.math.Transform;
import celery.engine.utils.math.Vec2F;
import celery.engine.utils.math.Vec3F;

public class GeometryData {

	protected Vec2F[][] points_;
	protected Vec3F[] circles_;
	protected Vec2F[][] edges_;
	protected Vec2F minCollision_ = new Vec2F();
	protected Vec2F normal = new Vec2F();

	public GeometryData(final Vec2F[][] points, final Vec3F[] cs) {
		setData(points, cs);
	}

	public GeometryData(final GeometryData gd1, final GeometryData gd2) {
		Vec2F[][] points = combine(gd1.points(), gd2.points());
		Vec3F[] circles = combine(gd1.circles(), gd2.circles());
		setData(points, circles);
	}

	public Vec2F minVectorOfCollision() {
		return minCollision_;
	}

	public void setVectorOfCollision(Vec2F v) {
		minCollision_ = v;
	}

	private final Vec3F[] combine(final Vec3F[] input1, final Vec3F[] input2) {
		final int length1 = input1 == null ? 0 : input1.length;
		final int length2 = input2 == null ? 0 : input2.length;
		final int total = length1 + length2;
		Vec3F[] out = new Vec3F[total];
		if (length1 != 0)
			System.arraycopy(input1, 0, out, 0, length1);
		if (length2 != 0)
			System.arraycopy(input2, 0, out, length1, length2);
		return out;
	}

	private final Vec2F[][] combine(final Vec2F[][] input1,
			final Vec2F[][] input2) {

		final int length1 = input1 == null ? 0 : input1.length;
		final int length2 = input2 == null ? 0 : input2.length;

		final int total = length1 + length2;

		Vec2F[][] out = new Vec2F[total][];

		int i = 0;

		for (int j = 0; j < length1; i++, j++) {
			out[i] = new Vec2F[input1[j].length];
			System.arraycopy(input1[j], 0, out[i], 0, input1[j].length);
		}

		for (int j = 0; j < length2; i++, j++) {
			out[i] = new Vec2F[input2[j].length];
			System.arraycopy(input2[j], 0, out[i], 0, input2[j].length);
		}

		return out;
	}

	public GeometryData transform(final Transform tfo, GeometryData out) {
		if (points_ != null && points_.length > 0) {
			final int nPolygons = points_.length;
			final Vec2F[][] pointsT = out.points();
			for (int i = 0; i < nPolygons; i++) {
				final int nPoints = pointsT[i].length;
				for (int j = 0; j < nPoints; j++) {
					pointsT[i][j].set(Utils.toWorldSpace(tfo, points_[i][j]));
				}
			}
		}

		if (circles_ != null && circles_.length > 0) {
			final int nVec3Fs = circles_.length;
			final Vec3F[] circlesT = out.circles();
			for (int i = 0; i < nVec3Fs; i++) {
				circlesT[i].x = tfo.translate.x + circles_[i].x;
				circlesT[i].y = tfo.translate.y + circles_[i].y;
				circlesT[i].z = tfo.scale.x * circles_[i].z;

			}
		}
		return out;
	}

	public final Vec2F[][] points() {
		return points_;
	}

	public final Vec3F[] circles() {
		return circles_;
	}

	public final Vec2F[][] edges() {
		if (points_ != null) {
			final int nPolygons = points_.length;
			for (int i = 0; i < nPolygons; i++) {
				final int nPoints = points_[i].length;
				for (int j = 0; j < nPoints; j++) {
					// edges_[i][j] = points_[i][(j + 1) % nPoints]
					// .sub(points_[i][j]);
					edges_[i][j].set(points_[i][(j + 1) % nPoints]).subEq(
							points_[i][j]);
				}
			}
		}
		return edges_;
	}

	/**
	 * 
	 * @param tfo
	 *            The transformation of the points/circles (usually the
	 *            primitive's transformation)
	 * @param points
	 *            A list of list of points of each polygon's points (may be
	 *            null)
	 * @param cs
	 *            A list of circles contained by the primitive (may be null)
	 */
	protected final void setData(Vec2F[][] points, Vec3F[] cs) {
		if (points != null) {
			final int nPolygons = points.length;

			// points_ = points;

			points_ = new Vec2F[nPolygons][];
			edges_ = new Vec2F[nPolygons][];
			for (int i = 0; i < nPolygons; i++) {
				final int nPoints = points[i].length;
				points_[i] = new Vec2F[nPoints];
				// System.arraycopy(points[i], 0, points_[i], 0, nPoints);
				edges_[i] = new Vec2F[nPoints];
				for (int g = 0; g < nPoints; g++) {
					points_[i][g] = new Vec2F(points[i][g]);
					edges_[i][g] = new Vec2F();
				}
			}
		}

		if (cs != null) {
			final int nVec3Fs = cs.length;
			if (nVec3Fs > 0) {
				circles_ = new Vec3F[nVec3Fs];
				for (int i = 0; i < nVec3Fs; i++) {
					circles_[i] = new Vec3F(cs[i]);
				}
				// System.arraycopy(cs, 0, circles_, 0, nVec3Fs);
			}
		}
	}

	public final boolean isCollision(final GeometryData d) {

		final Vec2F[][] pointsT = d.points();
		final Vec2F[][] edgesT = d.edges();
		final Vec2F[][] edgesS = this.edges();
		final Vec2F[][] pointsS = this.points();
		final Vec3F[] circlesT = d.circles();
		final Vec3F[] circlesS = this.circles();

		final int nPolygonsS = pointsS == null ? 0 : pointsS.length;
		final int nVec3FsS = circlesS == null ? 0 : circlesS.length;
		final int nPolygonsT = pointsT == null ? 0 : pointsT.length;
		final int nVec3FsT = circlesT == null ? 0 : circlesT.length;
		boolean isCollision = false;
		for (int i = 0; i < nPolygonsS; i++) {
			for (int j = 0; j < nPolygonsT; j++) { // polygon polygon collisions
				if (isPolyCollidePoly(pointsS[i], edgesS[i], pointsT[j],
						edgesT[j]))

					isCollision = true;
			}
			for (int j = 0; j < nVec3FsT; j++) { // polygon circle collisions

				if (isPolyCollideCircle(circlesT[j], pointsS[i], edgesS[i]))
					isCollision = true;
			}
		}

		for (int i = 0; i < nVec3FsS; i++) {
			for (int j = 0; j < nPolygonsT; j++) { // circle polygon collisions

				if (isPolyCollideCircle(circlesS[i], pointsT[j], edgesT[j]))
					isCollision = true;
			}
			for (int j = 0; j < nVec3FsT; j++) { // circle circle collisions

				if (isCircleCollideCircle(circlesS[i], circlesT[j]))
					isCollision = true;
			}
		}

		return isCollision;
	}

	protected Vec2F axis_ = new Vec2F(), axiss_ = new Vec2F(),
			tmp0_ = new Vec2F(), tmp1_ = new Vec2F(), tmp2_ = new Vec2F();

	private boolean isPolyCollidePoly(Vec2F[] verticesS, Vec2F[] edgesS,
			Vec2F[] verticesT, Vec2F[] edgesT) {

		/*
		 * For each edge in polygon S, project onto the perpendicular axis and
		 * determine the min or max
		 */
		boolean isColliding = true;

		for (int k = 0; k < edgesS.length; k++) {
			axis_.set(edgesS[k].y, -edgesS[k].x);
			axis_.normalize();
			float sMin = Float.POSITIVE_INFINITY, sMax = Float.NEGATIVE_INFINITY;

			for (int l = 0; l < verticesS.length; l++) {
				final float pos = verticesS[l].dot(axis_);
				sMin = Math.min(sMin, pos);
				sMax = Math.max(sMax, pos);
			}

			float tMin = Float.POSITIVE_INFINITY, tMax = Float.NEGATIVE_INFINITY;

			for (int l = 0; l < verticesT.length; l++) {
				final float pos = verticesT[l].dot(axis_);
				tMin = Math.min(tMin, pos);
				tMax = Math.max(tMax, pos);
			}
			if (sMax < tMin || tMax < sMin) {
				isColliding = false;
				break;
			}
		}
		if (!isColliding)
			return isColliding;

		for (int k = 0; k < edgesT.length; k++) {

			final Vec2F axis = new Vec2F(edgesT[k].y, -edgesT[k].x).normalize();

			float sMin = Float.POSITIVE_INFINITY, sMax = Float.NEGATIVE_INFINITY;

			for (int l = 0; l < verticesS.length; l++) {
				final float pos = verticesS[l].dot(axis);
				sMin = Math.min(sMin, pos);
				sMax = Math.max(sMax, pos);
			}

			float tMin = Float.POSITIVE_INFINITY, tMax = Float.NEGATIVE_INFINITY;

			for (int l = 0; l < verticesT.length; l++) {
				final float pos = verticesT[l].dot(axis);
				tMin = Math.min(tMin, pos);
				tMax = Math.max(tMax, pos);
			}
			if (sMax < tMin || tMax < sMin) {
				isColliding = false;
			}
		}

		return isColliding;
	}

	private boolean isPolyCollideCircle(Vec3F c, Vec2F[] verticesT,
			Vec2F[] edgesT) {
		Vec2F nearest = verticesT[0];
		float nearestD = verticesT[0].distance2(c);
		for (int i = 1; i < verticesT.length; i++) {
			float d = verticesT[i].distance2(c);
			if (d < nearestD) {
				nearest = verticesT[i];
				nearestD = d;
			}
		}
		// System.out.println(nearest);
		tmp2_.set(c.x, c.y);
		axis_.set(nearest).subEq(c.x, c.y).normalize();
		axiss_.set(axis_).scaleEq(c.z);
		tmp0_.set(tmp2_).addEq(axiss_);
		tmp1_.set(tmp2_).subEq(axiss_);
		final float p0 = tmp0_.dot(axis_);
		final float p1 = tmp1_.dot(axis_);
		final float pmin = Math.min(p0, p1);
		final float pmax = Math.max(p0, p1);
		float sMin = Float.POSITIVE_INFINITY;
		float sMax = Float.NEGATIVE_INFINITY;
		for (int i = 0; i < verticesT.length; i++) {
			final float proj = verticesT[i].dot(axis_);
			sMin = Math.min(sMin, proj);
			sMax = Math.max(sMax, proj);
		}
		if ((sMin < pmin && sMax < pmin) || (sMin > pmax && sMax > pmax)) { // if
			// not
			// overlap
			return false;
		}
		float minS = Float.POSITIVE_INFINITY;
		for (int i = 0; i < edgesT.length; i++) {
			axis_.set(edgesT[i].y, -edgesT[i].x).normalize();
			float tMin = Float.POSITIVE_INFINITY, tMax = Float.NEGATIVE_INFINITY;
			for (int l = 0; l < verticesT.length; l++) {
				final float pos = verticesT[l].dot(axis_);
				tMin = Math.min(tMin, pos);
				tMax = Math.max(tMax, pos);
			}
			axiss_.set(axis_).scaleEq(c.z);
			tmp0_.set(tmp2_).addEq(axiss_);
			tmp1_.set(tmp2_).subEq(axiss_);
			final float c0 = tmp0_.dot(axis_);
			final float c1 = tmp1_.dot(axis_);
			final float cmin = Math.min(c0, c1);
			final float cmax = Math.max(c0, c1);
			if ((tMin < cmin && tMax < cmin) || (tMin > cmax && tMax > cmax)) { // if
																				// not
				// overlap
				return false;
			} else { // determine amount of overlap
				float s = determineOverlapVector(tMin, tMax, cmin, cmax);
				if (Math.abs(s) < minS) {
					// axis_.set(nearest).subEq(c.x, c.y).normalize();
					minCollision_.set(axis_).scaleEq(-s);
					minS = Math.abs(s);
				}
			}
		}
		// float s = determineOverlapVector(sMin, sMax, pmin, pmax, axis);
		//
		// minCollision_ = axis.scale(-s);
		// System.out.println(s + ":::" + minCollision_ + ":::" + nearest);
		return true;
	}

	public float determineOverlapVector(float sMin, float sMax, float tMin,
			float tMax) {

		float scale = 1.f;
		// s0---t0---t1---s1
		if (sMin < tMin && tMax < sMax) {
			scale = Math.min(tMin - sMax, tMax - sMin);
		}
		// t0---s0---s1---t1
		else if (tMin < sMin && sMax < tMax) {
			scale = Math.min(tMin - sMax, tMax - sMin);
		}

		// t0---s0---t1---s1
		else if (tMin < sMin && tMax < sMax) {
			scale = tMax - sMin;
		}

		// s0---t0---s1---t1
		else if (sMin < tMin && sMax < tMax) {
			scale = tMin - sMax;
		}
		return scale;
	}

	private boolean isCircleCollideCircle(Vec3F c0, Vec3F c1) {
		float R2 = (c0.z + c1.z);
		R2 *= R2;
		float r2 = (c0.x - c1.x) * (c0.x - c1.x) + (c0.y - c1.y)
				* (c0.y - c1.y);
		axis_.set(c0.x, c0.y).subEq(c1.x, c1.y).normalize();

		axiss_.set(axis_).scaleEq(c0.z);
		tmp2_.set(c0.x, c0.y);
		tmp0_.set(tmp2_).addEq(axiss_);
		tmp1_.set(tmp2_).subEq(axiss_);
		final float p0 = tmp0_.dot(axis_);
		final float p1 = tmp1_.dot(axis_);

		axiss_.set(axis_).scaleEq(c1.z);
		tmp2_.set(c1.x, c1.y);
		tmp0_.set(tmp2_).addEq(axiss_);
		tmp1_.set(tmp2_).subEq(axiss_);
		final float p2 = tmp0_.dot(axis_);
		final float p3 = tmp1_.dot(axis_);

		float s = determineOverlapVector(Math.min(p0, p1), Math.max(p0, p1),
				Math.min(p2, p3), Math.max(p2, p3));
		minCollision_.set(axis_).scaleEq(s);
		return r2 <= R2;
	}
}
