package celery.engine.updatehandlers;

import android.content.Context;
import android.graphics.Canvas;
import android.graphics.Color;
import android.os.Message;
import android.os.SystemClock;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.Surface;
import android.view.SurfaceHolder;
import android.view.SurfaceView;
import celery.engine.Engine;
import celery.engine.graphics.Scene;
import celery.engine.graphics.Viewport;
import celery.engine.inputhandlers.TouchInputHandler;
import celery.engine.utils.MessageHandler;
import celery.engine.utils.math.Vec2F;

public final class DrawHandler extends SurfaceView implements UpdateHandler,
		SurfaceHolder.Callback {

	protected Surface surface_;
	protected final SurfaceHolder holder_ = this.getHolder();
	protected boolean isSurfaceCreated_ = false;
	protected DrawHandlerSettings settings_;
	protected long mSecFPSWait;
	protected Viewport viewport_;
	protected Vec2F canvasSize_;
	protected TouchInputHandler touchHandler_;

	public DrawHandler(DrawHandlerSettings settings, Context context) {
		super(context);
		settings_ = settings;

		mSecFPSWait = (int) (1000.f / settings.fps);
		this.getHolder().addCallback(this);
		setFocusable(true);
	}

	public void setTouchInputHandler(TouchInputHandler hdl) {
		touchHandler_ = hdl;
	}

	protected Scene scene_;

	public void setScene(Scene sc) {
		scene_ = sc;
	}

	public DrawHandlerSettings settings() {
		return settings_;
	}

	public Scene scene() {
		return scene_;
	}

	public void setViewport(Viewport view) {
		viewport_ = view;
		if (isSurfaceCreated_)
			viewport_.updateScreenSize(canvasSize_);
	}

	public Viewport viewport() {
		return viewport_;
	}

	public Vec2F canvasSize() {
		return canvasSize_;
	}

	@Override
	public void pause() {
		// TODO Auto-generated method stub

	}

	@Override
	public void start() {

		this.sendDrawMessage();
	}

	@Override
	public void resume() {

		this.sendDrawMessage();
	}

	@Override
	public void receiveMessage(Message msg) {
		this.draw();
		this.sendDrawMessage();
	}

	protected void sendDrawMessage() {
		MessageHandler h = Engine.instance().messageHandler();
		h.sendMessageAtTime(this, h.obtainMessage(), SystemClock.uptimeMillis()
				+ mSecFPSWait);
	}

	public void draw() {
		if (!Engine.instance().isRunning())
			return;

		if (!isSurfaceCreated_)
			return;
		Canvas canvas = null;

		try {
			canvas = holder_.lockCanvas();
			if (canvas == null)
				return;
			viewport_.updateScreenSize(canvasSize_);
			if (scene_ != null) {
				if (scene_.backgroundImage() == null) {
					canvas.drawColor(scene_.backgroundColor());
				} else {
					canvas.drawBitmap(scene_.backgroundImage(), 0.f, 0.f, null);
				}

				if (viewport_ != null) {
					scene_.draw(canvas, Engine.instance(), viewport_);
				}
			} else {
				canvas.drawColor(Color.DKGRAY);
			}

		} finally {
			if (canvas != null) {
				holder_.unlockCanvasAndPost(canvas);
			}
		}
	}

	@Override
	public void surfaceChanged(SurfaceHolder arg0, int arg1, int arg2, int arg3) {
		Canvas canvas = null;
		try {
			canvas = holder_.lockCanvas();
			canvasSize_ = new Vec2F(canvas.getWidth(), canvas.getHeight());
			if (viewport_ != null)
				viewport_.updateScreenSize(canvasSize_);
		} finally {
			if (canvas != null) {
				holder_.unlockCanvasAndPost(canvas);
			}
		}
	}

	@Override
	public void surfaceCreated(SurfaceHolder arg0) {
		isSurfaceCreated_ = true;
		Canvas canvas = null;
		try {
			canvas = holder_.lockCanvas();
			canvasSize_ = new Vec2F(canvas.getWidth(), canvas.getHeight());

			if (viewport_ != null)
				viewport_.updateScreenSize(canvasSize_);

		} finally {
			if (canvas != null) {
				holder_.unlockCanvasAndPost(canvas);
			}
		}
	}

	@Override
	public boolean onKeyDown(int keyCode, KeyEvent event) {
		if (Engine.instance().isRunning())
			Engine.instance().keyboardHandler().onKey(this, keyCode, event);
		return super.onKeyDown(keyCode, event);
	}

	@Override
	public boolean onKeyUp(int keyCode, KeyEvent event) {
		if (Engine.instance().isRunning())
			Engine.instance().keyboardHandler().onKey(this, keyCode, event);
		return super.onKeyUp(keyCode, event);
	}

	@Override
	public void surfaceDestroyed(SurfaceHolder arg0) {
		isSurfaceCreated_ = false;
	}

	@Override
	public void stop() {
		// TODO Auto-generated method stub

	}

	@Override
	public boolean onTouchEvent(MotionEvent event) {
		if (Engine.instance().isRunning()) {
			switch (event.getAction()) {
			case MotionEvent.ACTION_DOWN:
				Engine.instance().touchHandler().onTouch(null, this, event);
				break;
			case MotionEvent.ACTION_MOVE:
				Engine.instance().touchHandler().onDrag(null, this, event);
				break;
			case MotionEvent.ACTION_UP:
				Engine.instance().touchHandler().onRelease(null, this, event);
				break;
			}
		}
		return true;
	}

}
