package celery.engine.utils.math;

import android.util.FloatMath;
import celery.engine.utils.Utils;

/**
 * Defines a generic 2D vector and useful operations.
 * 
 * @author psastras
 * 
 */

public final class Vec2F {

	public float x, y;

	public Vec2F(float d0, float d1) {
		x = d0;
		y = d1;
	}

	public Vec2F(Vec2F v2f) {
		x = v2f.x;
		y = v2f.y;
	}

	public Vec2F(Vec3F v2f) {
		x = v2f.x;
		y = v2f.y;
	}

	public Vec2F() {
		x = 0.f;
		y = 0.f;
	}

	public final Vec2F add(Vec2F rhs) {
		return new Vec2F(this).addEq(rhs);
	}

	public final Vec2F addEq(Vec2F rhs) {
		x += rhs.x;
		y += rhs.y;
		return this;
	}

	public final Vec2F add(Vec3F rhs) {
		return new Vec2F(this).addEq(rhs);
	}

	public final Vec2F addEq(Vec3F rhs) {
		x += rhs.x;
		y += rhs.y;
		return this;
	}

	public Vec2F set(float xnew, float ynew) {
		x = xnew;
		y = ynew;
		return this;
	}

	public Vec2F set(final Vec2F vecnew) {
		x = vecnew.x;
		y = vecnew.y;
		return this;
	}

	public final Vec2F sub(Vec2F rhs) {
		return new Vec2F(this).subEq(rhs);
	}

	public final Vec2F sub(Vec3F rhs) {
		return new Vec2F(this).subEq(rhs);
	}

	public final Vec2F subEq(Vec2F rhs) {
		x -= rhs.x;
		y -= rhs.y;
		return this;
	}

	public final Vec2F subEq(float xn, float yn) {
		x -= xn;
		y -= yn;
		return this;
	}

	public final Vec2F subEq(Vec3F rhs) {
		x -= rhs.x;
		y -= rhs.y;
		return this;
	}

	public final Vec2F mul(Vec2F rhs) {
		return new Vec2F(this).mulEq(rhs);
	}

	public final Vec2F mulEq(Vec2F rhs) {
		x *= rhs.x;
		y *= rhs.y;
		return this;
	}

	public final Vec2F scale(float rhs) {
		return new Vec2F(this).scaleEq(rhs);
	}

	public final Vec2F scaleEq(float rhs) {
		x *= rhs;
		y *= rhs;
		return this;
	}

	public final Vec2F div(Vec2F rhs) {
		return new Vec2F(this).divEq(rhs);
	}

	public final Vec2F divEq(Vec2F rhs) {
		x /= rhs.x;
		y /= rhs.y;
		return this;
	}

	public final float dot(Vec2F rhs) {
		return x * rhs.x + y * rhs.y;
	}

	public final float dot(Vec3F rhs) {
		return x * rhs.x + y * rhs.y;
	}

	public final float length2() {
		return x * x + y * y;
	}

	public final float length() {
		return (float) Math.sqrt(this.length2());
	}

	public final float min() {
		return x < y ? x : y;
	}

	public final float max() {
		return x > y ? x : y;
	}

	public final boolean isEqual(Vec2F rhs) {
		return rhs.x == x && rhs.y == y;
	}

	public final float distance2(Vec2F rhs) {
		return (rhs.x - x) * (rhs.x - x) + (rhs.y - y) * (rhs.y - y);
	}

	public final float distance2(Vec3F rhs) {
		return (rhs.x - x) * (rhs.x - x) + (rhs.y - y) * (rhs.y - y);
	}

	public final float distance(Vec2F rhs) {
		return FloatMath.sqrt(this.distance2(rhs));
	}

	public final Vec2F rotateCWEq(final Vec2F center, float thetaDegrees) {
		final float theta = Utils.toRadians(thetaDegrees);
		this.subEq(center);
		float xtmp = x;
		float ytmp = y;
		x = (FloatMath.cos(theta) * xtmp + FloatMath.sin(theta) * ytmp);
		y = (-FloatMath.sin(theta) * xtmp + FloatMath.cos(theta) * ytmp);

		return this.addEq(center);
	}

	public final Vec2F rotateCW(Vec2F center, float thetaDegrees) {
		return new Vec2F(this).rotateCWEq(center, thetaDegrees);
	}

	public final Vec2F rotateCCWEq(Vec2F center, float thetaDegrees) {
		final float theta = thetaDegrees;// Utils.toRadians(thetaDegrees);
		Vec2F pT = this.sub(center);
		x = (FloatMath.cos(theta) * pT.x - FloatMath.sin(thetaDegrees) * pT.y);
		y = (FloatMath.sin(theta) * pT.x + FloatMath.cos(thetaDegrees) * pT.y);
		return this.addEq(center);
	}

	public final Vec2F rotateCCW(Vec2F center, float thetaDegrees) {
		return new Vec2F(this).rotateCCWEq(center, thetaDegrees);
	}

	public final Vec2F normalize() {
		return this.scaleEq(1.f / this.length());
	}

	public final Vec2F getNormalized() {
		return new Vec2F(this).normalize();
	}

	@Override
	public String toString() {
		return "[" + x + ", " + y + "]";
	}
}
